/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.alfa.bescheid;

import static java.util.Optional.*;
import static org.springframework.hateoas.server.mvc.WebMvcLinkBuilder.*;

import java.util.Optional;
import java.util.function.Predicate;

import org.apache.commons.collections.CollectionUtils;
import org.springframework.hateoas.CollectionModel;
import org.springframework.hateoas.EntityModel;
import org.springframework.hateoas.server.RepresentationModelAssembler;
import org.springframework.hateoas.server.mvc.WebMvcLinkBuilder;
import org.springframework.stereotype.Component;

import de.ozgcloud.alfa.bescheid.BescheidController.BescheidByVorgangController;
import de.ozgcloud.alfa.common.ModelBuilder;
import de.ozgcloud.alfa.common.binaryfile.BinaryFileController;
import de.ozgcloud.alfa.common.command.CommandController;
import de.ozgcloud.alfa.common.command.CommandController.CommandByRelationController;
import de.ozgcloud.alfa.postfach.PostfachMailController;
import de.ozgcloud.alfa.vorgang.Eingang;
import de.ozgcloud.alfa.vorgang.EingangHeader;
import de.ozgcloud.alfa.vorgang.VorgangController;
import de.ozgcloud.alfa.vorgang.VorgangHead;
import de.ozgcloud.alfa.vorgang.VorgangProperties;
import de.ozgcloud.alfa.vorgang.VorgangProperties.VorgangProperty;
import de.ozgcloud.alfa.vorgang.VorgangWithEingang;
import lombok.RequiredArgsConstructor;

@Component
@RequiredArgsConstructor
public class BescheidModelAssembler implements RepresentationModelAssembler<Bescheid, EntityModel<Bescheid>> {

	static final String REL_DELETE = "delete";
	static final String REL_UPLOAD_BESCHEID_FILE = "uploadBescheidFile";
	static final String REL_UPLOAD_ATTACHMENT = "uploadAttachment";
	static final String REL_ATTACHMENTS = "attachments";
	static final String REL_UPDATE = "update";
	static final String REL_CREATE_DOCUMENT = "createDocument";
	static final String REL_CREATE_DOCUMENT_FROM_FILE = "createDocumentFromFile";
	static final String REL_BESCHEID_DOCUMENT = "bescheidDocument";
	static final String REL_BESCHEIDEN_UND_SENDEN = "bescheidenUndSenden";
	static final String REL_BESCHEIDEN = "bescheiden";

	private static final Predicate<Bescheid> HAS_ATTACHMENTS = bescheid -> CollectionUtils.isNotEmpty(bescheid.getAttachments());

	private final VorgangController vorgangController;
	private final PostfachMailController postfachMailController;

	private final BescheidService bescheidService;
	private final VorgangProperties vorgangProperties;

	@Override
	public EntityModel<Bescheid> toModel(Bescheid bescheid) {
		var selfLink = linkTo(methodOn(BescheidController.class).getDraft(bescheid.getVorgangId()));
		var deleteLink = linkTo(
				methodOn(CommandByRelationController.class).createCommand(bescheid.getVorgangId(), bescheid.getId(), bescheid.getVersion(), null));
		var uploadBescheidFileLink = linkTo(methodOn(BinaryFileController.class).uploadFile(bescheid.getVorgangId(), "bescheidFile", null));
		var uploadAttachmentLink = linkTo(methodOn(BinaryFileController.class).uploadFile(bescheid.getVorgangId(), "bescheidAttachment", null));
		var attachmentsLink = linkTo(methodOn(BescheidController.class).getAttachments(bescheid.getId()));
		var createCommandLink = buildCreateCommandLink(bescheid);
		var vorgangWithEingang = vorgangController.getVorgang(bescheid.getVorgangId());
		var bescheidenUndSendenLink = linkTo(
				methodOn(BescheidCommandController.class).createCommand(vorgangWithEingang.getId(), bescheid.getId(), bescheid.getVersion(),
						null));

		return ModelBuilder.fromEntity(bescheid)
				.addLink(selfLink.withSelfRel())
				.addLink(deleteLink.withRel(REL_DELETE))
				.addLink(selfLink.withSelfRel())
				.addLink(uploadBescheidFileLink.withRel(REL_UPLOAD_BESCHEID_FILE))
				.addLink(uploadAttachmentLink.withRel(REL_UPLOAD_ATTACHMENT))
				.ifMatch(HAS_ATTACHMENTS)
				.addLink(attachmentsLink.withRel(REL_ATTACHMENTS))
				.addLink(createCommandLink.withRel(REL_UPDATE))
				.ifMatch(() -> canCreateBescheidDocumentAutomatically(vorgangWithEingang))
				.addLink(createCommandLink.withRel(REL_CREATE_DOCUMENT))
				.addLink(createCommandLink.withRel(REL_CREATE_DOCUMENT_FROM_FILE))
				.ifMatch(() -> canSendMessageToAntragsteller(vorgangWithEingang))
				.addLink(bescheidenUndSendenLink.withRel(REL_BESCHEIDEN_UND_SENDEN))
				.addLink(createCommandLink.withRel(REL_BESCHEIDEN))
				.buildModel();
	}

	@Override
	public CollectionModel<EntityModel<Bescheid>> toCollectionModel(Iterable<? extends Bescheid> entities) {
		return RepresentationModelAssembler.super.toCollectionModel(entities)
				.add(linkTo(BescheidByVorgangController.class).withSelfRel());
	}

	private WebMvcLinkBuilder buildCreateCommandLink(Bescheid bescheid) {
		return linkTo(methodOn(CommandController.CommandByRelationController.class).createCommand(bescheid.getVorgangId(), bescheid.getId(),
				bescheid.getVersion(), null));
	}

	boolean canCreateBescheidDocumentAutomatically(VorgangWithEingang vorgang) {
		return bescheidService.canCreateBescheidDocumentAutomatically() && hasVorgangCreateBescheidDocumentEnabled(vorgang);
	}

	boolean hasVorgangCreateBescheidDocumentEnabled(VorgangWithEingang vorgang) {
		return ofNullable(vorgang.getEingang())
				.map(Eingang::getHeader)
				.map(this::isCreateBescheidDocumentEnabled)
				.orElse(false);
	}

	private boolean isCreateBescheidDocumentEnabled(EingangHeader eingangHeader) {
		return vorgangProperties.getBescheid().stream().anyMatch(prop -> isFormIdAndFormEngineNameMatching(eingangHeader, prop));
	}

	private boolean isFormIdAndFormEngineNameMatching(EingangHeader eingangHeader, VorgangProperty property) {
		return property.getFormId().equals(eingangHeader.getFormId()) && property.getFormEngineName().equals(eingangHeader.getFormEngineName());
	}

	private boolean canSendMessageToAntragsteller(VorgangWithEingang vorgangWithEingang) {
		return postfachMailController.isPostfachConfigured() && hasServiceKonto(vorgangWithEingang);
	}

	private boolean hasServiceKonto(VorgangWithEingang vorgang) {
		return Optional.ofNullable(vorgang.getHeader()).map(VorgangHead::getServiceKonto).isPresent();
	}
}
