package de.ozgcloud.alfa.forwarding;

import static org.springframework.hateoas.server.mvc.WebMvcLinkBuilder.*;

import java.util.Optional;

import org.springframework.hateoas.EntityModel;
import org.springframework.hateoas.Link;
import org.springframework.hateoas.LinkRelation;
import org.springframework.hateoas.server.RepresentationModelProcessor;
import org.springframework.stereotype.Component;

import de.ozgcloud.alfa.common.FeatureToggleProperties;
import de.ozgcloud.alfa.common.command.CommandController;
import de.ozgcloud.alfa.common.user.CurrentUserService;
import de.ozgcloud.alfa.common.user.UserRole;
import de.ozgcloud.alfa.vorgang.Vorgang.VorgangStatus;
import de.ozgcloud.alfa.vorgang.VorgangWithEingang;
import lombok.RequiredArgsConstructor;

@Component
@RequiredArgsConstructor
class ForwardingVorgangWithEingangProcessor implements RepresentationModelProcessor<EntityModel<VorgangWithEingang>> {

	static final LinkRelation REL_FORWARD_BY_OZGCLOUD = LinkRelation.of("forwardByOzgCloud");
	static final LinkRelation REL_FORWARD_BY_EMAIL = LinkRelation.of("forwardByEmail");

	private final FeatureToggleProperties featureToggleProperties;
	private final CurrentUserService userService;

	@Override
	public EntityModel<VorgangWithEingang> process(EntityModel<VorgangWithEingang> model) {

		Optional.ofNullable(model.getContent())
				.ifPresent(vorgang -> model
						.addIf(isForwardableByOzgCloud(vorgang), () -> buildForwardByOzgCloudLink(vorgang))
						.addIf(isForwardableByEmail(), () -> buildForwardByEmailLink(vorgang)));
		return model;
	}

	boolean isForwardableByOzgCloud(VorgangWithEingang vorgang) {
		return featureToggleProperties.isForwardByOzgCloudEnabled() && isStatusNeu(vorgang) && isVerwaltungUser();
	}

	private boolean isStatusNeu(VorgangWithEingang vorgang) {
		return vorgang.getStatus() == VorgangStatus.NEU;
	}

	private boolean isVerwaltungUser() {
		return userService.hasRole(UserRole.VERWALTUNG_USER);
	}

	private Link buildForwardByOzgCloudLink(VorgangWithEingang vorgang) {
		return linkTo(methodOn(CommandController.CommandByRelationController.class).createCommand(vorgang.getId(), vorgang.getId(),
				vorgang.getVersion(), null)).withRel(REL_FORWARD_BY_OZGCLOUD);
	}

	boolean isForwardableByEmail() {
		return userService.hasRole(UserRole.EINHEITLICHER_ANSPRECHPARTNER);
	}

	private Link buildForwardByEmailLink(VorgangWithEingang vorgang) {
		return linkTo(methodOn(ForwardingController.class).findByVorgangId(vorgang.getId())).withRel(REL_FORWARD_BY_EMAIL);
	}

}
