package de.ozgcloud.alfa.wiedervorlage;

import java.util.Map;
import java.util.Optional;
import java.util.stream.Stream;

import org.apache.commons.collections.MapUtils;
import org.springframework.stereotype.Component;

import de.ozgcloud.alfa.common.command.Command;
import de.ozgcloud.alfa.common.command.CommandBodyMapper;
import de.ozgcloud.alfa.common.command.CommandOrder;
import de.ozgcloud.alfa.common.command.LegacyOrder;
import de.ozgcloud.alfa.historie.HistorieProcessor;

@Component
class WiedervorlageHistorieProcessor implements HistorieProcessor {

	@Override
	public Stream<Command> process(Stream<Command> commands) {
		return commands
				.map(this::translateCreateAttachedItem)
				.map(this::translateUpdateAttachedItem)
				.map(this::translatePatchAttachedItem);
	}

	Command translateCreateAttachedItem(Command command) {
		if (isCreateWiederVorlage(command)) {
			return command.toBuilder().order(LegacyOrder.CREATE_WIEDERVORLAGE).build();
		}
		return command;
	}

	private boolean isCreateWiederVorlage(Command command) {
		return CommandOrder.CREATE_ATTACHED_ITEM == command.getCommandOrder()
				&& isWiedervorlageItem(command);
	}

	Command translateUpdateAttachedItem(Command command) {
		if (isEditWiedervorlage(command)) {
			return command.toBuilder().order(LegacyOrder.EDIT_WIEDERVORLAGE).build();
		}
		return command;
	}

	private boolean isEditWiedervorlage(Command command) {
		return CommandOrder.UPDATE_ATTACHED_ITEM == command.getCommandOrder()
				&& isWiedervorlageItem(command);
	}

	Command translatePatchAttachedItem(Command command) {
		if (isPatchWiedervorlage(command)) {
			return translatePatchWiedervorlage(command);
		}
		return command;
	}

	private boolean isPatchWiedervorlage(Command command) {
		return CommandOrder.PATCH_ATTACHED_ITEM == command.getCommandOrder() && isWiedervorlageItem(command);
	}

	private boolean isWiedervorlageItem(Command command) {
		return getItemName(command).filter(name -> name.equals(Wiedervorlage.class.getSimpleName())).isPresent();
	}

	private Optional<String> getItemName(Command command) {
		return Optional.ofNullable(MapUtils.getString(command.getBody(), CommandBodyMapper.ITEM_NAME_PROPERTY));
	}

	Command translatePatchWiedervorlage(Command command) {
		return getDoneValue(command).map(isDone -> resolvePatchWiedervorlage(command, isDone)).orElse(command);

	}

	private Optional<Boolean> getDoneValue(Command command) {
		return Optional.ofNullable(MapUtils.getBoolean(getItemMap(command), Wiedervorlage.DONE_FIELD));
	}

	@SuppressWarnings("unchecked")
	private Map<String, Object> getItemMap(Command command) {
		return MapUtils.getMap(command.getBody(), CommandBodyMapper.ITEM_PROPERTY);
	}

	private Command resolvePatchWiedervorlage(Command command, Boolean isDone) {
		return isDone.booleanValue() ? command.toBuilder().order(LegacyOrder.WIEDERVORLAGE_ERLEDIGEN).build()
				: command.toBuilder().order(LegacyOrder.WIEDERVORLAGE_WIEDEREROEFFNEN).build();
	}
}
