package de.ozgcloud.alfa.loeschanforderung;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Collection;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.UUID;
import java.util.stream.Stream;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.EnumSource;
import org.junit.jupiter.params.provider.ValueSource;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.mockito.invocation.InvocationOnMock;

import com.thedeanda.lorem.LoremIpsum;

import de.ozgcloud.alfa.common.attacheditem.VorgangAttachedItem;
import de.ozgcloud.alfa.common.attacheditem.VorgangAttachedItemService;
import de.ozgcloud.alfa.common.command.Command;
import de.ozgcloud.alfa.common.command.CommandOrder;
import de.ozgcloud.alfa.common.command.CommandService;
import de.ozgcloud.alfa.common.command.CommandStatus;
import de.ozgcloud.alfa.common.command.CommandTestFactory;

class LoeschAnforderungHistorieProcessorTest {

	@Spy
	@InjectMocks
	private LoeschAnforderungHistorieProcessor processor;

	@Mock
	private CommandService commandService;
	@Mock
	private VorgangAttachedItemService vorgangAttachedItemService;

	@Nested
	class TestProcess {

		private final Command command = CommandTestFactory.create();

		@Test
		void shouldCallIsNotRevokedLoeschAnforderungZuruecknehmen() {
			process();

			verify(processor).isNotRevokedLoeschAnforderungZuruecknehmen(command);
		}

		@Nested
		class TestOnIsNotRevokedLoeschAnforderungZuruecknehmen {

			@BeforeEach
			void mock() {
				doReturn(true).when(processor).isNotRevokedLoeschAnforderungZuruecknehmen(any());
			}

			@Test
			void shouldCallIsNotCreateLoeschAnforderungVorgangAttachedItem() {
				process();

				verify(processor).isNotCreateLoeschAnforderungVorgangAttachedItem(command);
			}

			@Nested
			class TestOnIsNotCreateLoeschAnforderungVorgangAttachedItem {
				@BeforeEach
				void mock() {
					doReturn(true).when(processor).isNotCreateLoeschAnforderungVorgangAttachedItem(any());
				}

				@Test
				void shouldAddLoeschAnforderungZuruecknehmenRelatedCommandIds() {
					process();

					verify(processor).addLoeschAnforderungZuruecknehmenRelatedCommandIds(argThat(Collection::isEmpty), eq(command));
				}

				@Test
				void shouldRemoveLoeschAnforderungZuruecknehmenRelatedCommands() {
					doAnswer(this::addCommandIdToSet).when(processor)
							.addLoeschAnforderungZuruecknehmenRelatedCommandIds(anySet(), any());

					var commands = process();

					assertThat(commands).isEmpty();
				}

				@SuppressWarnings("unchecked")
				private Object addCommandIdToSet(InvocationOnMock invokation) {
					invokation.getArgument(0, Set.class).add(CommandTestFactory.ID);
					return command;
				}

				@Test
				void shouldKeepCommandsWithDifferentId() {
					doAnswer(this::addDummyIdToSet).when(processor)
							.addLoeschAnforderungZuruecknehmenRelatedCommandIds(anySet(), any());

					var commands = process();

					assertThat(commands).containsExactly(command);
				}

				@SuppressWarnings("unchecked")
				private Object addDummyIdToSet(InvocationOnMock invokation) {
					invokation.getArgument(0, Set.class).add("dummy");
					return command;
				}
			}

			@Nested
			class TestOnIsCreateLoeschAnforderungVorgangAttachedItem {
				@BeforeEach
				void mock() {
					doReturn(false).when(processor).isNotCreateLoeschAnforderungVorgangAttachedItem(any());
				}

				@Test
				void shouldNotCallAddLoeschAnforderungZuruecknehmenRelatedCommandIds() {
					process();

					verify(processor, never()).addLoeschAnforderungZuruecknehmenRelatedCommandIds(any(), any());
				}

				@Test
				void shouldRemoveCommand() {
					var commands = process();

					assertThat(commands).isEmpty();
				}
			}
		}

		@Nested
		class TestOnIsRevokedLoeschAnforderungZuruecknehmen {

			@BeforeEach
			void mock() {
				doReturn(false).when(processor).isNotRevokedLoeschAnforderungZuruecknehmen(any());
			}

			@Test
			void shouldNotGetLoeschAnforderungZuruecknehmenRelatedCommandIds() {
				process();

				verify(processor, never()).addLoeschAnforderungZuruecknehmenRelatedCommandIds(anySet(), any());
			}

			@Test
			void shouldRemoveRevokedLoeschAnforderungenZuruecknehmen() {
				var commands = process();

				assertThat(commands).isEmpty();
			}
		}

		private Stream<Command> process() {
			return processor.process(Stream.of(command));
		}

	}

	@Nested
	class TestIsNotRevokedLoeschAnforderungZuruecknehmen {

		@Nested
		class TestOnIsNotLoeschAnforderungZurücknehmen {

			private final Command command = CommandTestFactory.create();

			@Test
			void shouldReturnTrue() {
				var result = processor.isNotRevokedLoeschAnforderungZuruecknehmen(command);

				assertThat(result).isTrue();
			}
		}

		@Nested
		class TestOnIsLoeschAnforderungZurücknehmen {

			private final Command command = CommandTestFactory.createBuilder()
					.order(CommandOrder.LOESCH_ANFORDERUNG_ZURUECKNEHMEN.toString())
					.build();

			private final String changeStatusCommandId = UUID.randomUUID().toString();

			@BeforeEach
			void mock() {
				doReturn(changeStatusCommandId).when(processor).getChangeStatusCommandId(any());
				when(commandService.getById(any())).thenReturn(CommandTestFactory.create());
			}

			@Test
			void shouldGetChangeStatusCommandId() {
				isNotRevokedLoeschAnforderungZuruecknehmen();

				verify(processor).getChangeStatusCommandId(command);
			}

			@Test
			void shouldGetChangeStatusCommand() {
				isNotRevokedLoeschAnforderungZuruecknehmen();

				verify(commandService).getById(changeStatusCommandId);
			}

			@Test
			void shouldReturnTrueIfChangeStatusCommandIsNotRevoked() {
				var changeStatusCommand = CommandTestFactory.createBuilder().status(CommandStatus.FINISHED).build();
				when(commandService.getById(any())).thenReturn(changeStatusCommand);

				var result = isNotRevokedLoeschAnforderungZuruecknehmen();

				assertThat(result).isTrue();
			}

			@Test
			void shouldReturnFalseIfChangeStatusCommandIsRevoked() {
				var changeStatusCommand = CommandTestFactory.createBuilder().status(CommandStatus.REVOKED).build();
				when(commandService.getById(any())).thenReturn(changeStatusCommand);

				var result = isNotRevokedLoeschAnforderungZuruecknehmen();

				assertThat(result).isFalse();
			}

			private boolean isNotRevokedLoeschAnforderungZuruecknehmen() {
				return processor.isNotRevokedLoeschAnforderungZuruecknehmen(command);
			}
		}
	}

	@Nested
	class TestIsNotCreateLoeschAnforderungVorgangAttachedItem {

		@ParameterizedTest
		@EnumSource(names = { "CREATE_ATTACHED_ITEM" }, mode = EnumSource.Mode.EXCLUDE)
		void shouldReturnTrueForNotCreateAttachedItem(CommandOrder order) {
			var command = CommandTestFactory.createBuilder().order(order.name()).build();

			var result = processor.isNotCreateLoeschAnforderungVorgangAttachedItem(command);

			assertThat(result).isTrue();
		}

		@ParameterizedTest
		@EnumSource(names = { "CREATE_ATTACHED_ITEM" }, mode = EnumSource.Mode.EXCLUDE)
		void shouldNotCallVorgangAttachedItemService(CommandOrder order) {
			var command = CommandTestFactory.createBuilder().order(order.name()).build();

			processor.isNotCreateLoeschAnforderungVorgangAttachedItem(command);

			verify(vorgangAttachedItemService, never()).isLoeschAnforderung(any());
		}

		@Nested
		class TestOnCreateAttacheditemOrder {

			private final String itemName = LoremIpsum.getInstance().getWords(1);
			private final Map<String, ?> commandBody = Map.of(VorgangAttachedItem.FIELD_ITEM_NAME, itemName);
			private final Command command = CommandTestFactory.createBuilder()
					.order(CommandOrder.CREATE_ATTACHED_ITEM.name())
					.body(commandBody)
					.build();

			@Test
			void shouldCallVorgangAttachedItemService() {
				processor.isNotCreateLoeschAnforderungVorgangAttachedItem(command);

				verify(vorgangAttachedItemService).isLoeschAnforderung(itemName);
			}

			@ParameterizedTest
			@ValueSource(booleans = { true, false })
			void shouldReturnNegatedResult(boolean isLoeschAnforderung) {
				when(vorgangAttachedItemService.isLoeschAnforderung(any())).thenReturn(isLoeschAnforderung);

				var result = processor.isNotCreateLoeschAnforderungVorgangAttachedItem(command);

				assertThat(result).isNotEqualTo(isLoeschAnforderung);
			}
		}
	}

	@Nested
	class TestGetChangeStatusCommandId {

		@Test
		void shouldReturnChangeStatusCommandId() {
			var changeStatusCommandId = UUID.randomUUID().toString();
			var commandBody = Map.of(DeleteLoeschAnforderung.CHANGE_STATUS_COMMAND_ID_FIELD, changeStatusCommandId);
			var command = CommandTestFactory.createBuilder().body(commandBody).build();

			var id = processor.getChangeStatusCommandId(command);

			assertThat(id).isEqualTo(changeStatusCommandId);
		}
	}

	@Nested
	class TestAddLoeschAnforderungZuruecknehmenRelatedCommandIds {
		private final String changeStatusCommandId = UUID.randomUUID().toString();
		private final String deleteAttachedItemCommandId = UUID.randomUUID().toString();
		private final Map<String, ?> commandBody = Map.of(
				DeleteLoeschAnforderung.CHANGE_STATUS_COMMAND_ID_FIELD, changeStatusCommandId,
				DeleteLoeschAnforderung.DELETE_ATTACHED_ITEM_COMMAND_ID_FIELD, deleteAttachedItemCommandId,
				LoremIpsum.getInstance().getWords(1), UUID.randomUUID().toString());
		private final Command command = CommandTestFactory.createBuilder()
				.order(CommandOrder.LOESCH_ANFORDERUNG_ZURUECKNEHMEN.name())
				.body(commandBody)
				.build();
		private final HashSet<String> idsToRemove = new HashSet<>();

		@Test
		void shouldReturnSameCommand() {
			var returnedCommand = processor.addLoeschAnforderungZuruecknehmenRelatedCommandIds(idsToRemove, command);

			assertThat(returnedCommand).isSameAs(command);
		}

		@Test
		void shouldNotAddAnyIds() {
			processor.addLoeschAnforderungZuruecknehmenRelatedCommandIds(idsToRemove, CommandTestFactory.create());

			assertThat(idsToRemove).isEmpty();
		}

		@Test
		void shouldAddRelatedCommandIdsOfLoeschAnforderungenZuruecknehmen() {
			processor.addLoeschAnforderungZuruecknehmenRelatedCommandIds(idsToRemove, command);

			assertThat(idsToRemove).containsExactlyInAnyOrder(changeStatusCommandId, deleteAttachedItemCommandId);
		}
	}

}
