/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.alfa.loeschanforderung;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;

import de.ozgcloud.alfa.common.command.CommandOrder;
import de.ozgcloud.alfa.common.user.CurrentUserService;
import de.ozgcloud.alfa.common.user.UserRole;

class LoeschAnforderungPreAuthorizeHandlerTest {

	@InjectMocks
	private LoeschAnforderungPreAuthorizeHandler handler;

	@Mock
	private LoeschAnforderungService loeschAnforderungService;
	@Mock
	private CurrentUserService currentUserService;

	@Nested
	class TestHasPermissionForOrder {

		private final LoeschAnforderung loeschAnforderung = LoeschAnforderungTestFactory.create();

		@DisplayName("Has permission for VORGANG_LOESCHEN")
		@Nested
		class TestVorgangLoeschen {

			@BeforeEach
			void init() {
				when(loeschAnforderungService.getById(LoeschAnforderungTestFactory.ID)).thenReturn(loeschAnforderung);
			}

			@Test
			void shouldLoadLoeschAnforderung() {
				handler.hasPermissionForOrder(LoeschAnforderungTestFactory.ID, CommandOrder.VORGANG_LOESCHEN);

				verify(loeschAnforderungService).getById(LoeschAnforderungTestFactory.ID);
			}

			@Test
			void shouldCallLoeschAnforderungService() {
				handler.hasPermissionForOrder(LoeschAnforderungTestFactory.ID, CommandOrder.VORGANG_LOESCHEN);

				verify(loeschAnforderungService).isAllowedToVorgangLoeschen(loeschAnforderung);
			}

			@Test
			void shouldReturnFalseIfNotAllowed() {
				when(loeschAnforderungService.isAllowedToVorgangLoeschen(loeschAnforderung)).thenReturn(false);

				var result = handler.hasPermissionForOrder(LoeschAnforderungTestFactory.ID, CommandOrder.VORGANG_LOESCHEN);

				assertThat(result).isFalse();
			}

			@Test
			void shouldReturnTrueIfAllowed() {
				when(loeschAnforderungService.isAllowedToVorgangLoeschen(loeschAnforderung)).thenReturn(true);

				var result = handler.hasPermissionForOrder(LoeschAnforderungTestFactory.ID, CommandOrder.VORGANG_LOESCHEN);

				assertThat(result).isTrue();
			}
		}

		@DisplayName("Has permission for LOESCH_ANFORDERUNG_ZURUECKNEHMEN")
		@Nested
		class TestVorgangLoeschenZuruecknehmen {

			@Nested
			class checkingRole {
				@BeforeEach
				void setupUserServiceMock() {
					when(currentUserService.hasRole(any())).thenReturn(false);
				}

				@Test
				void shouldReturnTrueForVerwaltungUser() {
					when(currentUserService.hasRole(UserRole.VERWALTUNG_USER)).thenReturn(true);

					var result = handler.hasPermissionForOrder(LoeschAnforderungTestFactory.ID, CommandOrder.LOESCH_ANFORDERUNG_ZURUECKNEHMEN);

					assertThat(result).isTrue();
				}

				@Test
				void shouldReturnTrueForEA() {
					when(currentUserService.hasRole(UserRole.EINHEITLICHER_ANSPRECHPARTNER)).thenReturn(true);

					var result = handler.hasPermissionForOrder(LoeschAnforderungTestFactory.ID, CommandOrder.LOESCH_ANFORDERUNG_ZURUECKNEHMEN);

					assertThat(result).isTrue();
				}

				@Test
				void shouldReturnFalseMissingRole() {
					when(currentUserService.hasRole(UserRole.VERWALTUNG_USER)).thenReturn(false);

					var result = handler.hasPermissionForOrder(LoeschAnforderungTestFactory.ID, CommandOrder.LOESCH_ANFORDERUNG_ZURUECKNEHMEN);

					assertThat(result).isFalse();
				}
			}

			@Test
			void shouldReturnFalseForNotAllowedCommandOrder() {
				var result = handler.hasPermissionForOrder(LoeschAnforderungTestFactory.ID, CommandOrder.CREATE_ATTACHED_ITEM);

				assertThat(result).isFalse();
			}
		}
	}
}