/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.alfa.postfach;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.UUID;
import java.util.stream.Stream;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.hateoas.EntityModel;
import org.springframework.hateoas.IanaLinkRelations;
import org.springframework.hateoas.Link;
import org.springframework.hateoas.RepresentationModel;
import org.springframework.hateoas.mediatype.hal.HalModelBuilder;
import org.springframework.web.util.UriComponentsBuilder;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.datatype.jdk8.Jdk8Module;
import com.fasterxml.jackson.datatype.jsr310.JavaTimeModule;

import de.ozgcloud.alfa.common.user.UserId;
import de.ozgcloud.alfa.common.user.UserManagerUrlProvider;
import de.ozgcloud.alfa.postfach.PostfachMail.Direction;
import de.ozgcloud.alfa.postfach.PostfachMailController.PostfachMailCommandByVorgangController;
import de.ozgcloud.alfa.vorgang.ServiceKontoTestFactory;
import de.ozgcloud.alfa.vorgang.VorgangController;
import de.ozgcloud.alfa.vorgang.VorgangHeadTestFactory;
import de.ozgcloud.alfa.vorgang.VorgangHeaderTestFactory;
import de.ozgcloud.alfa.vorgang.VorgangWithEingang;
import de.ozgcloud.alfa.vorgang.VorgangWithEingangTestFactory;
import lombok.SneakyThrows;

class PostfachMailModelAssemblerTest {

	@InjectMocks
	@Spy
	private PostfachMailModelAssembler modelAssembler;
	@Mock
	private UserManagerUrlProvider userManagerUrlProvider;
	@Mock
	private VorgangController vorgangController;

	@Nested
	class TestToModel {

		@BeforeEach
		void mock() {
			when(userManagerUrlProvider.isConfiguredForUserProfile()).thenReturn(false);
		}

		@Test
		void shouldHaveSelfLink() {
			var link = modelAssembler.toModel(PostfachMailTestFactory.create()).getLink(IanaLinkRelations.SELF_VALUE);

			assertThat(link).isPresent().get().extracting(Link::getHref).isEqualTo(PostfachMailController.PATH + "/" + PostfachMailTestFactory.ID);
		}

		@Nested
		class ResendLink {

			@DisplayName("if 'sentSuccessful' is false, add link")
			@Test
			void shouldBePresent() {
				var link = modelAssembler.toModel(PostfachMailTestFactory.createBuilder().sentSuccessful(false).build())
						.getLink(PostfachMailModelAssembler.REL_RESEND_POSTFACH_MAIL);

				assertThat(link).isPresent().get().extracting(Link::getHref)
						.isEqualTo("/api/postfachMails/" + PostfachMailTestFactory.ID + "/commands");
			}

			@DisplayName("if 'sentSuccessful' is true, hide link")
			@Test
			void shouldNotBePresentIfSuccessfulSent() {
				var link = modelAssembler.toModel(PostfachMailTestFactory.createBuilder().sentSuccessful(true).build())
						.getLink(PostfachMailModelAssembler.REL_RESEND_POSTFACH_MAIL);

				assertThat(link).isNotPresent();
			}

			@Test
			void shouldNotBePresentIfNotYetTriedToSend() {
				var nachricht = PostfachMailTestFactory.createBuilder().sentSuccessful(false).sentAt(null).build();

				var link = modelAssembler.toModel(nachricht).getLink(PostfachMailModelAssembler.REL_RESEND_POSTFACH_MAIL);

				assertThat(link).isNotPresent();
			}

		}

		@Nested
		class SendLink {

			@DisplayName("should be present if nachricht not jet tried to send")
			@Test
			void shouldBePresent() {
				var nachricht = PostfachMailTestFactory.createBuilder().sentSuccessful(false).sentAt(null).build();

				var link = modelAssembler.toModel(nachricht).getLink(PostfachMailModelAssembler.REL_SEND);

				assertThat(link).isPresent().get().extracting(Link::getHref)
						.isEqualTo("/api/postfachMails/" + PostfachMailTestFactory.ID + "/commands");
			}

			@Test
			void shouldNotBePresentIfNachrichtSent() {
				var nachricht = PostfachMailTestFactory.createBuilder().sentSuccessful(true).build();

				assertThat(modelAssembler.toModel(nachricht).getLink(PostfachMailModelAssembler.REL_SEND)).isNotPresent();
			}

		}

		@Nested
		class AttachmentsLink {

			@Test
			@DisplayName("should be present if attachments present")
			void shouldBePresent() {
				var link = modelAssembler.toModel(PostfachMailTestFactory.create()).getLink(PostfachMailModelAssembler.REL_ATTACHMENTS);

				assertThat(link).isPresent().get().extracting(Link::getHref)
						.isEqualTo("/api/postfachMails/" + PostfachMailTestFactory.ID + "/attachments");
			}

			@Test
			@DisplayName("should NOT be present if there are no attachments")
			void shouldNotBePresentOnNoAttachments() {
				var link = modelAssembler.toModel(PostfachMailTestFactory.createBuilder().clearAttachments().build())
						.getLink(PostfachMailModelAssembler.REL_ATTACHMENTS);

				assertThat(link).isNotPresent();
			}
		}

		@DisplayName("created by link")
		@Nested
		class CreatedByLink {

			@Nested
			class TestOnConfiguredUserManager {

				private final String userProfileTemplate = "UserProfileTemplateDummy/%s";

				@BeforeEach
				void mock() {
					when(userManagerUrlProvider.isConfiguredForUserProfile()).thenReturn(true);
				}

				@Test
				void shouldBePresent() {
					when(userManagerUrlProvider.getUserProfileTemplate()).thenReturn(userProfileTemplate);

					var link = modelAssembler.toModel(PostfachMailTestFactory.create()).getLink(PostfachMailModelAssembler.REL_CREATED_BY);

					assertThat(link).isPresent().get().extracting(Link::getHref)
							.isEqualTo("UserProfileTemplateDummy/" + PostfachMailTestFactory.CREATED_BY);
				}

				@DisplayName("should not be present if the value of createdBy starts with 'system'")
				@Test
				void shouldNotBePresentOnSystemUser() {
					var link = modelAssembler
							.toModel(PostfachMailTestFactory.createBuilder()
									.createdBy(UserId.from(PostfachMailModelAssembler.SYSTEM_USER_IDENTIFIER + UUID.randomUUID().toString())).build())
							.getLink(PostfachMailModelAssembler.REL_CREATED_BY);

					assertThat(link).isNotPresent();
				}

				@Test
				void shouldNotBePresentOnNull() {
					var link = modelAssembler.toModel(PostfachMailTestFactory.createBuilder().createdBy(null).build())
							.getLink(PostfachMailModelAssembler.REL_CREATED_BY);

					assertThat(link).isNotPresent();
				}
			}

			@Nested
			class TestOnNonConfiguredUserManager {

				@BeforeEach
				void mock() {
					when(userManagerUrlProvider.isConfiguredForUserProfile()).thenReturn(false);
				}

				@Test
				void shouldNOTBePresentOnNotConfiguredUserManager() {
					var link = modelAssembler
							.toModel(PostfachMailTestFactory.createBuilder()
									.createdBy(UserId.from(PostfachMailModelAssembler.SYSTEM_USER_IDENTIFIER + UUID.randomUUID().toString())).build())
							.getLink(PostfachMailModelAssembler.REL_CREATED_BY);

					assertThat(link).isNotPresent();
				}

				@Test
				void shouldNotCallUserManagerUrlProvider() {
					modelAssembler
							.toModel(PostfachMailTestFactory.createBuilder()
									.createdBy(UserId.from(PostfachMailModelAssembler.SYSTEM_USER_IDENTIFIER + UUID.randomUUID().toString())).build())
							.getLink(PostfachMailModelAssembler.REL_CREATED_BY);

					verify(userManagerUrlProvider, never()).getUserProfileTemplate();
				}

				@Test
				void shouldNotBePresentOnNull() {
					var link = modelAssembler.toModel(PostfachMailTestFactory.createBuilder().createdBy(null).build())
							.getLink(PostfachMailModelAssembler.REL_CREATED_BY);

					assertThat(link).isNotPresent();
				}
			}
		}

		@DisplayName("edit link")
		@Nested
		class TestEditLink {
			PostfachMail.PostfachMailBuilder UNSENT_MAIL_BUILDER = PostfachMailTestFactory.createBuilder()
					.direction(Direction.OUT).sentAt(null).sentSuccessful(false);

			@Test
			void shouldBePresentOnUnsendNachricht() {
				var postfachNachricht = UNSENT_MAIL_BUILDER.build();

				var link = modelAssembler.toModel(postfachNachricht).getLink(PostfachMailModelAssembler.REL_EDIT);

				assertThat(link).isPresent().get().extracting(Link::getHref)
						.isEqualTo("/api/postfachMails/" + PostfachMailTestFactory.ID + "/commands");
			}

			@Nested
			class ShouldNotBePresent {

				@Test
				void OnIncomingNachricht() {
					var postfachNachricht = UNSENT_MAIL_BUILDER.direction(Direction.IN).build();

					var link = modelAssembler.toModel(postfachNachricht).getLink(PostfachMailModelAssembler.REL_EDIT);

					assertThat(link).isNotPresent();
				}

				@Test
				void OnSentNachricht() {
					var postfachNachricht = UNSENT_MAIL_BUILDER.sentSuccessful(true).build();

					var link = modelAssembler.toModel(postfachNachricht).getLink(PostfachMailModelAssembler.REL_EDIT);

					assertThat(link).isNotPresent();
				}
			}
		}
	}

	@Nested
	class TestToCollectionModel {

		private final Stream<PostfachMail> mails = Stream.of(PostfachMailTestFactory.create());
		private final VorgangWithEingang vorgang = VorgangWithEingangTestFactory.create();
		private final PostfachSettings postfachSettings = PostfachSettingsTestFactory.create();

		@Mock
		private RepresentationModel<EntityModel<PostfachSettings>> model;

		@BeforeEach
		void setUpMocks() {
			doReturn(model).when(modelAssembler).buildHalRepresentationModel(mails, vorgang, postfachSettings);
		}

		@Test
		void shouldBuildHalRepresentationModel() {
			callModelAssembler();

			verify(modelAssembler).buildHalRepresentationModel(mails, vorgang, postfachSettings);
		}

		@Nested
		class OnHasServiceKonto {

			@Test
			void shouldAddPostfachNachrichtLinks() {
				doReturn(true).when(modelAssembler).hasServiceKonto(vorgang);

				callModelAssembler();

				verify(modelAssembler).addPostfachNachrichtLinks(model, vorgang);
			}
		}

		@Nested
		class OnHasNotServiceKonto {

			@Test
			void shouldAddPostfachNachrichtLinks() {
				doReturn(false).when(modelAssembler).hasServiceKonto(vorgang);

				callModelAssembler();

				verify(modelAssembler, never()).addPostfachNachrichtLinks(any(), any());
			}
		}

		@Test
		void shouldReturnModel() {
			var returnedModel = callModelAssembler();

			assertThat(returnedModel).isEqualTo(model);
		}

		private RepresentationModel<EntityModel<PostfachSettings>> callModelAssembler() {
			return modelAssembler.toCollectionModel(mails, vorgang, postfachSettings);
		}
	}

	@Nested
	class TestBuildHalRepresentationModel {

		private final PostfachMail postfachMail = PostfachMailTestFactory.create();
		private final Stream<PostfachMail> mails = Stream.of(postfachMail);
		private final VorgangWithEingang vorgang = VorgangWithEingangTestFactory.create();
		private final PostfachSettings postfachSettings = PostfachSettingsTestFactory.create();

		@Test
		void shouldHavePostfachAsContent() {
			var model = (EntityModel<PostfachSettings>) callModelAssembler();

			assertThat(model.getContent()).isEqualTo(postfachSettings);
		}

		@Test
		void shouldAddSelfLink() {
			var model = callModelAssembler();
			var selfLink = model.getLink(IanaLinkRelations.SELF_VALUE);

			assertThat(selfLink).get().extracting(Link::getHref).isEqualTo(
					String.format("%s?%s=%s", PostfachMailController.PATH, PostfachMailController.PARAM_VORGANG_ID, VorgangHeaderTestFactory.ID));
		}

		@Test
		void shouldCallToModel() {
			callModelAssembler();

			verify(modelAssembler).toModel(postfachMail);
		}

		@Test
		@SneakyThrows
		void shouldContainPostfachMails() {
			var objectMapper = new ObjectMapper().registerModules(new JavaTimeModule(), new Jdk8Module());
			doReturn(EntityModel.of(postfachMail)).when(modelAssembler).toModel(postfachMail);

			var model = callModelAssembler();

			assertThat(objectMapper.writeValueAsString(model)).contains(objectMapper.writeValueAsString(EntityModel.of(postfachMail)));
		}

		private RepresentationModel<EntityModel<PostfachSettings>> callModelAssembler() {
			return modelAssembler.buildHalRepresentationModel(mails, vorgang, postfachSettings);
		}

	}

	@Nested
	class TestHasServiceKonto {

		@Test
		void shouldReturnTrueIfServicekontoIsPresent() {
			var hasServiceKonto = modelAssembler.hasServiceKonto(VorgangWithEingangTestFactory.createBuilder()
					.header(VorgangHeadTestFactory.createBuilder().serviceKonto(ServiceKontoTestFactory.create()).build()).build());

			assertThat(hasServiceKonto).isTrue();
		}

		@Test
		void shouldReturnFalseIfServicekontoIsNotPresent() {
			var hasServiceKonto = modelAssembler.hasServiceKonto(VorgangWithEingangTestFactory.createBuilder()
					.header(VorgangHeadTestFactory.createBuilder().serviceKonto(null).build()).build());

			assertThat(hasServiceKonto).isFalse();
		}

		@Test
		void shouldReturnFalseOnMissingVorgangHeader() {
			var hasServiceKonto = modelAssembler.hasServiceKonto(VorgangWithEingangTestFactory.createBuilder()
					.header(null).build());

			assertThat(hasServiceKonto).isFalse();
		}
	}

	@Nested
	class TestAddPostfachNachrichtLinks {

		private final PostfachSettings postfach = PostfachSettingsTestFactory.create();

		private RepresentationModel<EntityModel<PostfachSettings>> model = HalModelBuilder
				.halModelOf(postfach)
				.build();

		@Nested
		class ResetNewPostfachNachrichtLink {
			@Test
			void shouldBePresent() {
				callModelAssembler(VorgangWithEingangTestFactory.create());

				var link = model.getLink(PostfachMailModelAssembler.REL_RESET_NEW_POSTFACH_MAIL);
				assertThat(link).isPresent().get().extracting(Link::getHref)
						.isEqualTo(UriComponentsBuilder.fromUriString("/api/vorgangs")
								.pathSegment(VorgangHeaderTestFactory.ID, "hasNewPostfachNachricht")
								.build().toString());
			}

			@Test
			void shouldNotBePresent() {
				callModelAssembler(VorgangWithEingangTestFactory.createBuilder().hasNewPostfachNachricht(false).build());

				var link = model.getLink(PostfachMailModelAssembler.REL_RESET_NEW_POSTFACH_MAIL);
				assertThat(link).isNotPresent();
			}
		}

		@DisplayName("send postfach mail")
		@Nested
		class SendPostfachMailLink {

			private final String sendPostfachMailLinkRel = PostfachMailModelAssembler.REL_SEND_POSTFACH_MAIL;
			private final VorgangWithEingang vorgang = VorgangWithEingangTestFactory.create();

			@Test
			void shouldCheckIfVorgangIsEditable() {
				callModelAssembler(vorgang);

				verify(vorgangController).isEditable(vorgang);
			}

			@Test
			void shouldExist() {
				when(vorgangController.isEditable(vorgang)).thenReturn(true);

				callModelAssembler(vorgang);

				var link = model.getLink(sendPostfachMailLinkRel);
				assertThat(link).get().extracting(Link::getHref)
						.isEqualTo(PostfachMailCommandByVorgangController.PATH.replace("{vorgangId}", VorgangHeaderTestFactory.ID));
			}

			@Test
			void shouldNotExistWhenVorgangIsNotEditable() {
				when(vorgangController.isEditable(vorgang)).thenReturn(false);

				callModelAssembler(vorgang);

				var link = model.getLink(sendPostfachMailLinkRel);
				assertThat(link).isEmpty();
			}
		}

		@Nested
		class UploadAttachmentLink {

			private final String uploadAttachmentLinkRel = PostfachMailModelAssembler.REL_UPLOAD_ATTACHMENT;

			@Test
			void shouldBePresent() {
				callModelAssembler(VorgangWithEingangTestFactory.create());

				var link = model.getLink(uploadAttachmentLinkRel);
				assertThat(link).get().extracting(Link::getHref)
						.isEqualTo(String.format("/api/binaryFiles/%s/postfachNachrichtAttachment/file", VorgangHeaderTestFactory.ID));
			}
		}

		private void callModelAssembler(VorgangWithEingang vorgang) {
			modelAssembler.addPostfachNachrichtLinks(model, vorgang);
		}
	}
}
