/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.alfa.postfach;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.io.File;
import java.io.FileOutputStream;
import java.io.OutputStream;
import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.util.Collections;
import java.util.List;
import java.util.stream.Stream;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.boot.test.mock.mockito.MockBean;

import de.ozgcloud.alfa.common.user.UserId;
import de.ozgcloud.alfa.common.user.UserProfileTestFactory;
import de.ozgcloud.alfa.common.user.UserService;
import de.ozgcloud.alfa.vorgang.AntragstellerTestFactory;
import de.ozgcloud.alfa.vorgang.EingangTestFactory;
import de.ozgcloud.alfa.vorgang.VorgangWithEingang;
import de.ozgcloud.alfa.vorgang.VorgangWithEingangTestFactory;
import lombok.SneakyThrows;

@SpringBootTest
class PostfachNachrichtPdfServiceITCase {

	@Autowired
	private PostfachNachrichtPdfService service;
	@MockBean
	private UserService userService;

	@DisplayName("Generate pdf file")
	@Nested
	class TestGeneratePdfFile {

		@BeforeEach
		void mock() {
			when(userService.getById(any(UserId.class))).thenReturn(UserProfileTestFactory.create());
		}

		@SneakyThrows
		@Test
		void generatePdfFile() {
			var tempFile = createTempFile();

			getAllAsPdf(VorgangWithEingangTestFactory.create(), new FileOutputStream(tempFile));

			assertThat(tempFile).isNotEmpty();
		}

		@SneakyThrows
		@Test
		void generatePdfFileAntragstellerNotSet() {
			var tempFile = createTempFile();

			getAllAsPdf(buildVorgangAntragstellerNotSet(), new FileOutputStream(tempFile));

			assertThat(tempFile).isNotEmpty();
		}

		@SneakyThrows
		@Test
		void generatePdfFileNoAttachments() {
			var tempFile = createTempFile();

			getAllAsPdf(buildVorgangAntragstellerNotSet(),
					Stream.of(PostfachNachrichtPdfDataTestFactory.createBuilder().attachmentNames(Collections.emptyList()).build()),
					new FileOutputStream(tempFile));

			assertThat(tempFile).isNotEmpty();
		}

		@SneakyThrows
		private File createTempFile() {
			var tempFile = File.createTempFile("ozgcloud_nachricht_", ".pdf");
			// tempFile.deleteOnExit();
			return tempFile;
		}

		@SneakyThrows
		private void getAllAsPdf(VorgangWithEingang vorgang, OutputStream out) {
			getAllAsPdf(vorgang, buildPostfachMails(), out);
		}

		@SneakyThrows
		private void getAllAsPdf(VorgangWithEingang vorgang, Stream<PostfachNachrichtPdfData> postfachNachrichten, OutputStream out) {
			service.getAllAsPdf(vorgang, postfachNachrichten, out);
			out.close();
		}
	}

	@DisplayName("Map Postfach-Nachricht")
	@Nested
	class TestMapPostfachNachricht {

		private final ZonedDateTime createdAt = ZonedDateTime.of(2023, 9, 25, 10, 0, 0, 0, ZoneId.of("UTC"));
		private final PostfachNachrichtPdfData pdfData = PostfachNachrichtPdfDataTestFactory.createBuilder().createdAt(createdAt).build();

		@Test
		void shouldMapCreatedAtWithCorrectTimezoneAndFormatting() {
			var nachricht = service.mapPostfachNachricht(pdfData, AntragstellerTestFactory.create());

			assertThat(nachricht.getCreatedAt()).isEqualTo("25.09.2023 12:00:00 MESZ");
		}
	}

	private VorgangWithEingang buildVorgangAntragstellerNotSet() {
		return VorgangWithEingangTestFactory.createBuilder().eingang(EingangTestFactory.createBuilder().antragsteller(null).build()).build();
	}

	private Stream<PostfachNachrichtPdfData> buildPostfachMails() {
		return Stream.of(PostfachNachrichtPdfDataTestFactory.createBuilder().subject("hase")
						.attachmentNames(List.of("Hase.png", "Hasenlied.mscz", "Erweitertes-Führungszeugniß.pdf", "Haftbefehl.pdf"))
						.build(),
				PostfachNachrichtPdfDataTestFactory.create(),
				PostfachNachrichtPdfDataTestFactory.create(),
				PostfachNachrichtPdfDataTestFactory.create(),
				PostfachNachrichtPdfDataTestFactory.create());
	}
}